import SOMA.agent.*;
import SOMA.naming.*;
import SOMA.resourceManagement.*;

import java.io.PrintStream;
import java.util.*;


/** Agent Monitoring Agent
  * Questo agente raccoglie i consumi e il numero di allarmi prodotti da un agente monitorato 
  * in determinati place; gli argomenti sono, nell'ordine, l'agentID e l'elenco dei place.
  */
public class AgentMA extends Agent{
  
  AgentInfo info = new AgentInfo();
  int cpuAlarms=0, memAlarms=0, netAlarms=0, fileAlarms=0;
  AgentID agID; 
  
  PlaceID initPlace = null, places[] = null;
  HashSet notReacheablePlaces = new HashSet();
  int index = -1;

  // si potrebbero aggiungere controlli, magari in assenza di place indicati esplicitamente
  // prendere come elenco di default quello di tutti i place del dominio
  public void putArgument( Object obj){
  
    Object[] args = (Object[])obj;
    try{ 
      agID = new AgentID((String)args[0]);  
      // ci sono problemi con agentID composti da pi di 2 stringhe, 
      // "nomeDominio nomePlace numeroProgressivo"; rivedere il parsing...
    }
    catch(NameException ex){
      ex.printStackTrace();
    }
    places = new PlaceID[args.length-1];

    for (int i=1; i<places.length; i++){
     try{
       places[i] = new PlaceID((String)args[i]);
     }
     catch(NameException ex){
       try{
         places[i] = new PlaceID("DummyDomain DummyPlace");
       }
       catch(NameException e){
         e.printStackTrace();
       }
     }
    }
  }


  public void run(){
    
    PrintStream out = agentSystem.getOut();  
    initPlace = agentSystem.getPlaceID();

    out.println("AgentMA -> visiter i seguenti places:");
    for (int i=0; i<places.length; i++)
      out.println((i+1) + ") " + places[i]);

    monitoringTour();
  }


  public void monitoringTour(){
    
    PrintStream out = agentSystem.getOut();
    index++;

    if (index<places.length){
     try{
       go(places[index], "getInfo"); 
     }
     catch( CantGoException ex){
      notReacheablePlaces.add(places[index]);
      ex.printStackTrace(out);
      out.println("AgentMA -> il place " + places[index]
              + " non  raggiungibile; vado al place successivo");
      monitoringTour();
     }
    }
    else{
      try{
        out.println("AgentMA -> tour di monitoraggio terminato, torno al place di partenza: " + initPlace);
        go(initPlace, "deliverInfo");
      }
      catch( CantGoException ex){
        ex.printStackTrace(out);
        out.println("AgentMA -> ritorno al place di partenza fallito");
      }
    }
  }


  public void getInfo()  {
    
    Enumeration en;
    // recupero dal PRM i consumi totali di tutti i thread dell'agente monitorato
    AgentInfo[] temp =(agentSystem.getEnvironment()).placeResourceManager.getAgentTotConsumptions(agID);
    
    // sommo i consumi dei vari thread dell'agente monitorato
    for (int i=0; i<temp.length;i++){
       info.time += temp[i].time;
       info.mem += temp[i].mem;
       info.file_in += temp[i].file_in;
   	   info.file_out += temp[i].file_out;
	     info.tcp_in += temp[i].tcp_in;
	     info.tcp_out += temp[i].tcp_out;
	     info.udp_in += temp[i].udp_in;
       info.udp_out += temp[i].udp_out;
     }
    
    // scandisco la tabella degli allarmi sulla cpu contando quelli prodotti dall'agente monitorato
    en = ((agentSystem.getEnvironment()).placeResourceManager.getCpuAlarmInfo()).keys();
  
    for (; en.hasMoreElements();)
     if ((AgentID)(en.nextElement())==agID)
       cpuAlarms+=1;
	
	  // scandisco la tabella degli allarmi sulla memoria contando quelli prodotti dall'agente monitorato
    en = ((agentSystem.getEnvironment()).placeResourceManager.getMemAlarmInfo()).keys();
  
    for (; en.hasMoreElements();)
     if ((AgentID)(en.nextElement())==agID)
       memAlarms+=1; 
	  
	  // scandisco la tabella degli allarmi sui file contando quelli prodotti dall'agente monitorato
    en = ((agentSystem.getEnvironment()).placeResourceManager.getFileAlarmInfo()).keys();
  
    for (; en.hasMoreElements();)
     if ((AgentID)(en.nextElement())==agID)
       fileAlarms+=1;  
    
    // scandisco la tabella degli allarmi sulla rete contando quelli prodotti dall'agente monitorato
    en = ((agentSystem.getEnvironment()).placeResourceManager.getNetAlarmInfo()).keys();
  
    for (; en.hasMoreElements();)
     if ((AgentID)(en.nextElement())==agID)
       netAlarms+=1; 
  

    monitoringTour();

  }



  public void deliverInfo(){
   
   PrintStream out = agentSystem.getOut();  
   out.println("AgentMA -> informazioni sull'agente monitorato:"+agID);
   out.println("Consumi totali raggiunti da tutti i suoi thread nel dominio:");
   info.print(out);
   out.println("Allarmi prodotti:");
   out.println("cpuAlarms = "+ cpuAlarms);
   out.println("memAlarms = "+ memAlarms);
   out.println("fileAlarms = "+ fileAlarms);
   out.println("netAlarms = "+ netAlarms);
  
  /* (agentSystem.getEnvironment()).domainResourceManager
                  .agentXConsumptions = new AgentInfo(info);
   (agentSystem.getEnvironment()).domainResourceManager
                  .agentXcpuAlarms = cpuAlarms;
   (agentSystem.getEnvironment()).domainResourceManager
                  .agentXmemAlarms = memAlarms;
   (agentSystem.getEnvironment()).domainResourceManager
                  .agentXfileAlarms = fileAlarms;
   (agentSystem.getEnvironment()).domainResourceManager
                  .agentXnetAlarms = netAlarms;
     
  */
  }
}
 