package MobileMonitoring;

import StaticMonitoring.Settings;
import SOMA.agent.*;
import SOMA.naming.*;
import SOMA.utility.*;
import com.adventnet.snmp.snmp2.*;
import java.io.*;
import java.util.*;
import java.net.*;

/**Agente Mobile che ricopre il ruolo di Manager Snmp ed ha la capacit
//di interrogare gli Agenti statici Snmp presenti sui place che visita;
// in grado di effettuare semplici interrogazioni Snmp: GET/GETNEXT/GETBULK
*/
public class SnmpAtomicOpMA extends SnmpV3MobileManager
{
    //lista degli OID oggetto della richiesta
    String[] reqOIDs = null;
    //tipo di operazione da effettuare su ogni Agente Snmp
    byte op_snmp;
    //eventuali argomenti propri dell'operazione richiesta (vd. nonrep e maxrep di getbulk)
    String[] argsop = null;

    /**qui si recuperano le singole informazioni a partire dall'oggetto passato
    //in input.
    //o e' un oggetto della classe InfoAtomicSnmpOp
    */
    public void putArgument(Object o)
    {
        InfoAtomicSnmpOp Info = (InfoAtomicSnmpOp) o;
        super.putArgument(Info);
        reqOIDs = Info.getIDs();
        op_snmp = Info.getOp();
        argsop = Info.getExtra();
    }

    /**questo  il metodo di partenza, eseguito all'attivazione dell'agente:
    //qui si avvia il cronometro che conter il tempo necessario al completamento
    //dell'operazione, si definir la prima tappa del "viaggio" e si inizializzer
    //il campo home che dovr poi essere utilizzato per il ritorno al Place
    //di partenza
    */
    public void run()
    {
        super.run();
    }

    /**metodo di passaggio ad una tappa successiva,  qui che si controlla se il viaggio
    // finito, in tal caso si torna al Place di partenza
    */
    public void goNext()
    {
        super.goNext();
    }

    //qui sono racchiuse le operazioni necessarie all'adempimento del compito
    //specifico di questo Agente Mobile
    public void doLocalJob(Settings sets,SnmpSession session)
    {
        PrintStream out = agentSystem.getOut();
        SnmpPDU answer = snmpOp(sets,session,reqOIDs,op_snmp,argsop);
        if (check(answer) == false)
        {
            result_code[current] = false;
            job_result.add(current,null);
            if (++current != Ntargets)
	            nexthop = targets[current];
            goNext();
        }
        Vector varbinds = answer.getVariableBindings();
        SnmpVar[] localres = new SnmpVar[varbinds.size()];
        int j=0;
        //dal vettore di var bind ricavo le snmpvar e le copio nel vettore dei risultati
        for (Enumeration e=varbinds.elements(); e.hasMoreElements(); j++)
            localres[j] = ((SnmpVarBind)e.nextElement()).getVariable();
        result_code[current] = true;
        job_result.add(current,localres);
        session.close();
        api.close();
        if (++current != Ntargets)
        	nexthop = targets[current];
        goNext();
    }

    //verifica se la PDU ottenuta in risposta contiene errori o  null
    boolean check(SnmpPDU pdu)
    {
        PrintStream out = agentSystem.getOut();
        //verifico se la pdu e' nulla (time out)
        if (pdu == null)
        {
            trace_err(out,getHeader()+" NULL PDU in response");
            return false;
        }

        //verifico se la pdu contiene errori
        if (pdu.getErrstat() > 0)
        {
            trace_err(out,getHeader()+" Error occurred in PDU "+pdu.getError());
			pdu.fix();
            return false;
        }

        return true;
    }


     SnmpPDU snmpOp(Settings sets,SnmpSession session,String OIDs[],byte opcode,String argsOp[])
     {
        PrintStream out = agentSystem.getOut();
        //creo una nuova PDU che dovro' poi spedire
        SnmpPDU pdu = new SnmpPDU();
        //setto il tipo di operazione nella pdu
        pdu.setCommand(opcode);
        if (opcode == SnmpAPI.GETBULK_REQ_MSG)
        {
            if (session.getVersion() == SnmpAPI.SNMP_VERSION_1)
            {
                trace_err(out,getHeader()+" Snmpv1 doesn't support GetBulk");
                return null;
            }
            pdu.setErrstat(Integer.parseInt(argsOp[0]));
            pdu.setErrindex(Integer.parseInt(argsOp[1]));
        }

        //introduco tutti gli OID richiesti nella PDU
        for (int j=0; j < OIDs.length; j++)
        {
            SnmpOID oid = new SnmpOID(OIDs[j]);
            if (oid.toValue() == null)
            {
                trace_err(out,getHeader()+" Invalid OID argument: "+OIDs[j]);
                return null;
            }
            else pdu.addNull(oid);
        }
		
		pdu.setUserName(sets.userName.getBytes());
		if (session.getVersion() == SnmpAPI.SNMP_VERSION_3)
		{
			if (sets.config == false)
			USMUtils.init_v3_params(sets.userName,sets.authProto,sets.authPwd,sets.privPwd,session.getPeername(),session.getRemotePort(),session);
			pdu.setContextName(sets.contextName.getBytes());
			pdu.setContextID(sets.contextID.getBytes());
		}
        
        try  { pdu = session.syncSend(pdu); }
        catch (SnmpException e)
        {
            trace_err(out,getHeader()+" Sending PDU"+e.getMessage());
            return null;
        }

        return(pdu);
   }

    //metodo di ritorno al Place di partenza
    public void goHome()
    {
        super.goHome();
    }

    //metodo che racchiude le azioni da svolgere una volta tornati al Place di partenza
    public Message EndofPlan()
    {
        return(super.EndofPlan());
    }

    //metodo di visualizazione dei risultati su un particolare PrintStream
    void ShowJob(PrintStream o)
    {
        o.println("Data retrieved from "+getID().toString());
        for (int j=0; j<Ntargets; j++)
        {
            o.println(targets[j].toString());
            for (int k=0; k<reqOIDs.length; k++)
                o.println(reqOIDs[k]+": "+((SnmpVar[])job_result.elementAt(j))[k].toTagString());
        }
    }

}

