package StaticMonitoring;

import MonitorConfig.*;
import java.util.*;
import java.io.PrintStream;
import com.adventnet.snmp.snmp2.*;

class ReceiverTableReq extends Debug implements SnmpClient,Operations
{
	//variabile che indica se il lavoro e' concluso
	protected boolean done;
	//variabile che indica se il lavoro si e' concluso senza errori
	boolean noerr;
	//numero di righe reperite
	int nrow;
	//numero di colonne che ci si deve aspettare
	int ncol;
	//variabile destinata a contenere i risultati
	snmpTable table = null;
	String oid_tab;
	//manager di riferimento
	StaticManager manager;
	Settings sets = null;
	//variabile in cui memorizzo l'identificatore dell'ultima richiesta in sospeso
	//N.B. visto che inoltro una richiesta solo quando mi arriva una risposta (cioe' nella callback)
	//avro' sempre al piu' una richiesta in sospeso
	int outstandID;
	//identificatore dell'ultima richiesta inviata
	int recentID;
	//ultima pdu inviata, puo' essere  utile in caso di timeout per poter
	//avviare una richiesta successiva
	SnmpPDU recentPDU;

	protected ReceiverTableReq(StaticManager man,snmpTable t,int firstID,boolean deb,PrintStream o)
	{
		super(deb,o);
		outstandID = firstID;
		manager = man;
		table = t;
		nrow = 0;
		ncol = t.getNCols();
		oid_tab = t.getTabOID();
		done = false;
		noerr = false;

	}

	public void initparams(Settings s)
	{
		sets = s;
	}
	
	public boolean callback(SnmpSession session,SnmpPDU pdu,int requestID)
	{
	System.out.println(requestID+" : received");
    	String[] OIDs = new String[ncol];
    	if (outstandID != requestID)
    	{
    		trace_err("Answer received "+requestID+" not expected ("+outstandID+")");
    		return true;
    	}

		//se tutto ok avvio la richiesta di un'altra riga e aggiungo quella ricevuta alla tabella
		boolean next = manager.check(pdu,oid_tab);
        if (next)
        {
        	table.addRow(pdu);
           	nrow++;
	        for(int i=0; i<ncol; i++)
	            OIDs[i] = pdu.getObjectID(i).toString();
	        //N.B anche la richiesta successiva e' asincrona quindi c'e'
	        //un ritorno immediato
	        int req = 0; //lascio che sia il sistema ad assegnare l'identificatore
	        //l'invocazione di snmpAsyncOp avviene passando settaggi nulli perche' sono stati gia' fatti
	        //nella prima richiesta
	        recentPDU = manager.snmpAsyncOp(sets,session,OIDs,GETNEXT,null,req);
	        recentID = recentPDU.getReqid();
	        //aggiungo l'identificatore della richiesta appena inviata alla lista
	        //di quelle in sospeso
	        outstandID = recentID;
       	}
	    else
	    {
	    	System.out.println("DONE: "+requestID);
	     //errore o risposta non piu' relativa alla tabella
            done = true;
      }
       	return next;
    }

    public boolean authenticate(SnmpPDU pdu,String community)
	{
		return true;
	}

	public void debugPrint(String debugOutput)
	{
	}

	//controlla se e' scaduto il timeout sulla richiesta correntemente in sospeso
	//restituisce true se e' scaduto il timeout sull'ultima richiesta fatta
	//			  false altrimenti (possibilita' che arrivi la risposta)
	public boolean checkTimeout(SnmpSession session)
	{
		//se sull'ultima richiesta e' scaduto il timeout devo avviare un'altra
		//richiesta a partire dall'ultima pdu inviata perche' comunque potrei
		//trovare altre righe
		if (session.checkTimeout(outstandID))
		{
		   /* String[] OIDs = new String[ncol];

			for(int i=0; i<ncol; i++)
	            OIDs[i] = recentPDU.getObjectID(i).toString();

	        int req = 0; //lascio che sia il sistema ad assegnare l'identificatore
	        recentPDU = manager.snmpAsyncOp(session,OIDs,GETNEXT,null,req);
	        recentID = recentPDU.getReqid();
	        //aggiungo l'identificatore della richiesta appena inviata alla lista
	        //di quelle in sospeso
	        outstandID = recentID;*/
	        done = true;
	        return true;
	     }
	     return false;
	}

	public boolean hasDone()
	{
		return done;
	}

	public snmpTable getAnswer()
	{
		return table;
	}
}