package StaticMonitoring;

import MonitorConfig.*;
import SOMA.*;

import SOMA.agent.*;
import SOMA.naming.*;
import SOMA.naming.place.*;
import SOMA.utility.Chronometer;
import java.io.*;
import java.util.*;
import java.net.*;
import com.adventnet.snmp.snmp2.*;
import com.adventnet.snmp.mibs.*;

public class StaticManager implements Operations
{
    //variabili che indicano se si e' in modalita' di debugging (output per errori)
    //e quali PrintStream utilizzare
    boolean debug = false;
    PrintStream out;

    SnmpAPI api;

    Configuration conf = null;

    SnmpSession discoverySession;

    //cronometro per il calcolo del tempo impiegato nello svolgere le operazioni Snmp
    Chronometer Crono = new Chronometer("Total Time");

	public StaticManager()
    {
    }

    public StaticManager(Configuration config,boolean deser,boolean deb,PrintStream o)
    {
    	conf = config;
        startEngine(deser,deb,o);
    }

    void trace_err(String s)
    {
        String current_class = (getClass().getName())+">";
        if (debug)
            out.println(current_class+s);
    }


	void trace(String s)
	{
		out.println(s);
	}

    //avvia il motore SNMP locale
    //deser e' una variabile che mi indica se l'avvio deve avvenire deserializzando la
    //configurazione V3 cioe' ristabilendo una configurazione definita in precedenza
    //deb e' lo stato del debugging: se true si genera output per il debugging
    //o e' il PrintStream di output
    void startEngine(boolean deser,boolean deb,PrintStream o)
    {
        debug = deb;
        out = o;

        api = new SnmpAPI();
        api.start();
        discoverySession = new SnmpSession(api);
        discoverySession.setVersion(SnmpAPI.SNMP_VERSION_3);
        try {discoverySession.open();}
        catch (SnmpException e)
        {
            trace_err("Error opening discovery session: "+e.getMessage());
            return ;
        }

        conf.startConfig(api,discoverySession,deser);
    }

    //esegue alcune inizializzazioni necessarie al successivo funzionamento
    //recupera la tabella dei motori Snmp e la tabella degli utenti
    //crea una tabella SupportedNodesMIB destinata a contenere le associazioni:
    //(IP:porta) --> (moduli MIB supportati) ed il nome della directory in cui si
    //trovano i MIB files
    //crea una sessione V3 che verra' utilizzata tutte le volte che sara' necessario
    //scoprire il nome di un motore Snmp (procedura di Discovery V3)
    //void init()
    //{
    //}


    //ferma il motore SNMP locale
    //se e' stato definito un file di serializzazione serializza la configurazione
    //V3 corrente
    public void stopEngine()
    {
        conf.stopConfig();
        discoverySession.close();
        api.close();
    }

    public void setDebug(boolean status)
    {
        debug = status;
        //n.b. forse da togliere e mettere a livello di MonitorTEMP
        conf.setDebug(status);
    }

    public boolean isDebug()
    {
        return debug;
    }

    public PrintStream getOut()
    {
        return out;
    }

    public void setOut(PrintStream ps)
    {
        out = ps;
    }

    public Configuration getCurrentConf()
    {
        return conf;
    }

    /* mi permette di scoprire quale versione supporta un agente remoto. Utile se si vuole un manager multilingue, che utilizza
    // lo stesso standard adottato dall'agente remoto cui si rivolge
    //host e' l'indirizzo dell'Agente Snmp remoto
    //port e' la porta d'ascolto dell'Agente Snmp remoto
    //restituisce la versione supportata dall'Agente Snmp remoto o -1 se non e' nessuna di quelle previste
    //(v1/v2c/v3) o se ci sono stati problemi (es. porta remota non corretta)
    */
    public int discoverAgentVersion(String host,int port)
    {
        //uso sysDescr perche' c'e' sempre..
        SnmpOID oid = new SnmpOID(".1.3.6.1.2.1.1.1.0");
        SnmpSession session = new SnmpSession(api);
        session.setTimeout(4000);
        session.setPeername(host);
        session.setRemotePort(port);
        session.setRetries(2);
        try
        {
            session.open();
        }
        catch(SnmpException e) {}

        try
        {
            session.setVersion(SnmpAPI.SNMP_VERSION_3);
            if (conf.engineTable.getEntry(host,port) == null)
            {
               SnmpEngineEntry snmpEntry = new SnmpEngineEntry(host,port);
               snmpEntry.discoverEngineID(discoverySession);
               conf.engineTable.addEntry(snmpEntry);
            }

            session.setUserName(new String("initial").getBytes());
            if (session.get(oid) != null)
                return(3);
        }
        catch(SnmpException ex){trace_err(ex.toString()); }

        try
        {
            session.setVersion(SnmpAPI.SNMP_VERSION_2C);
            if (session.get(oid) != null)
                return(2);
        }
        catch(SnmpException ex) {trace_err(ex.toString());  }

        try
        {
            session.setVersion(SnmpAPI.SNMP_VERSION_1);
            if (session.get(oid) != null)
                return(1);
        }
        catch(SnmpException ex){trace_err(ex.toString());  }

        session.close();
        return -1;

    }


   /*apre una sessione Snmp settandone i parametri conformemente a quanto specficato nei
   //parametri di input
   //host e' il nome dell'host su cui si trova l'agente Snmp da interrogare
   //args e' il vettore dei parametri: la posizione di ogni parametro e' definita dalle variabili
   //static della classe Settings
   //from_config stabilisce se i parametri sono tutti contenuti in args (false) o se devono anche essere recuperati
   //dalla configurazione del motore Snmp locale (true)
   //session e' la sessione Snmp gia' creata, ma su cui devono essere settati i parametri
   //restituisce la sessione aperta pronta per l'utilizzo e l'oggetto contenente tutti i settaggi
   */
   Settings prepareSession(String host,String args[],boolean from_config,SnmpSession session)
   {
        Settings sets = null;
        //se ci si appoggia alla configurazione, si devono recuperare altri parametri solo
        //in caso di V3
        /*if ( (from_config) && (args[Settings.VERSION].equals("v3")) )
		{
            sets = new Settings(session,host,args,conf.api,debug,out);
        	try { session.open(); }
	        catch (SnmpException ex)
    	    {
        	    trace_err("Problems in opening the session");
            	return null;
        	}
		}
        else
        {
        	sets = new Settings(session,host,args,debug,out);
        	try { session.open(); }
	        catch (SnmpException ex)
    	    {
        	    trace_err("Problems in opening the session");
            	return null;
        	}
        }*/
        if ( (from_config) && (args[Settings.VERSION].equals("v3")) )
			sets = new Settings(session,host,args,true,debug,out);
		else 
			sets = new Settings(session,host,args,false,debug,out);
			
        try { session.open(); }
	    catch (SnmpException ex)
    	{
            trace_err("Problems in opening the session");
           	return null;
        }
		
        if (sets.isSuccess() == false)
        {
            trace_err("Problems in creating a session with parameters:");
            trace_err(host);
            sets.showParams(args);
            return null;
        }

	    return sets;
   }

    /** prepara sessioni asincrone per l'interazione con ognuno degli agenti Snmp individuati da
    *   host e porta, considera i parametri comuni presenti in args (quelli opzionali: contextname,contextID,timeout,retries)
	*   e recupera quelli propri di ognuno (versione,parametri di sicurezza) dalla configurazione
	*   associa ad ogni sessione un ascoltatore SnmpClient
	*   hosts e' l'elenco degli IP su cui si trovano i vari agenti Snmp
	*   ports e' l'elenco delle porte di ascolto degli agenti Snmp
	*   args e' il vettore degli argomenti che inizialmente conterra' solo i parametri opzionali comuni a tutte
	*   le sessioni nelle posizioni definite dalle variabili static della classe Settings
	*   receivers e' l'elenco degli SnmpClient da associare come ascoltatori ad ognuna delle sessioni
	*   separate_thread indica se e' necessario associare un thread separato alla ricezione (true) o
	*   se sara' lo stesso thread che gestisce la sessione ad occuparsene
	*   restituisce l'elenco delle sessioni asincrone aperte e preparate
	*/
    Settings[] prepareAsyncSessions(SnmpSession[] sessions,String[] hosts,int[] ports,String[] args,SnmpClient[] receivers,boolean separate_thread)
    {
   		out.println(hosts.length+" "+ports.length);
    	boolean from_config = true;
    	Settings sets[] = new Settings[hosts.length];
    	
    	//qui preparo tutte le sessioni e creo i ricevitori asincroni
   		for(int k=0;k<hosts.length;k++)
   		{
   		out.println(hosts[k]+" "+ports[k]);
   			String host = hosts[k];
   			//introduco la porta d'ascolto dell'agente sul place tra i parametri
   			int port = ports[k];

			if (prepareArguments(host,args) == false)
				return null;
			
        	//setto i parametri di sessione conformemente alle informazioni che ho e
        	//recuperando le rimanenti (caso v3) dalla configurazione
        	sets[k] = prepareSession(host,args,from_config,sessions[k]);
        	//registro il gestore delle risposte come client Snmp della sessione
        	sessions[k].addSnmpClient(receivers[k]);
        	if (separate_thread)
        		sessions[k].setCallbackthread(true);
        }
        
        return sets;
    }

    //riempie il vettore extra_in contenente gia' eventuali parametri opzionali (porta,contextname,contexid,timeout,retries)
    //con i parametri associati al place dalla configurazione (versione,nome di sicurezza)
    //extra_in e' il vettore dei parametri: la posizione di ogni parametro e' definita dalle variabili
    //static della classe Settings, inizialmente potra' contenere solo quelli corrispondenti ad opzioni che non sono
    //parte della configurazione
    boolean prepareArguments(String host,String[] extra_in)
    {
    	int port;

    	if (extra_in[Settings.PORT] == null)
    	{
    		extra_in[Settings.PORT] = new String("161");
    	   	port = 161;
    	}
    	else
        	//nb qui ricavo la porta pero' deve essere stata correttamente settata
        	port = Integer.parseInt(extra_in[Settings.PORT]);

		//recupero versione supportata dall'agente Snmp, se non esiste un'entry corrispondente
		//segnalo l'errore ed esco
        int version = conf.agentSnmp.getSupportedVersion(host,port);
        if (version == -1)
        {
   			trace_err("SnmpAgent corresponding to: ("+host+":"+port+") not configured");
   			return false;
   		}

		//recupero la versione, se arrivo qui deve esistere un'entry
        if (version == SnmpAPI.SNMP_VERSION_1)
            extra_in[Settings.VERSION] = "v1";
        else if (version == SnmpAPI.SNMP_VERSION_2C)
            extra_in[Settings.VERSION] = "v2";
        else
            extra_in[Settings.VERSION] = "v3";

        String name =  conf.agentSnmp.getSecurityName(host,port);
        extra_in[Settings.SECNAME] = name;
        
        return true;
   }

   /*esegue una richiesta Get per 1 o piu' variabili a partire dai parametri specificati in
   //input: ciascuno occupa una precisa posizione all'interno di args[]
   //apre, usa e chiude un'apposita sessione Snmp
   //host e' l'indirizzo presso cui si trova l'agente Snmp da interrogare
   //args e' il vettore dei parametri: la posizione di ogni parametro e' definita dalle variabili
   //static della classe Settings
   //OIDs e' il vettore contenente tutti gli OID che si vogliono richiedere
   //from_config stabilisce se i parametri sono tutti contenuti in args (false) o se devono anche essere recuperati
   //dalla configurazione del motore Snmp locale (true)
   //restituisce la PDU di risposta o null se ci sono stati problemi
   */
	public SnmpPDU snmpGet(String host,String args[],String OIDs[], boolean from_config)
    {
    	SnmpSession session = new SnmpSession(api);
    	//se si richiede di ricorrere ai dati di configurazione si recuperano gli argomenti mancanti
    	//altrimenti e' gia' tutto in args
    	if (from_config)
    		if (prepareArguments(host,args) == false)
    			return null;

        //con tutti gli argomenti che ho preparo la sessione facendo gli opportuni settaggi
        Settings sets = prepareSession(host,args,from_config,session);
        SnmpPDU ris = snmpOp(sets,session,OIDs,GET,null);
        session.close();
        return ris;
    }


   /*esegue una richiesta Get per 1 o piu' variabili a partire dai parametri specificati in
   //input: ciascuno occupa una precisa posizione all'interno di args[]
   //apre, usa e chiude un'apposita sessione Snmp
   //placeID e' l'identificatore del place su cui si trova l'agente Snmp da interrogare
   //args e' il vettore dei parametri: la posizione di ogni parametro e' definita dalle variabili
   //static della classe Settings
   //OIDs e' il vettore contenente tutti gli OID che si vogliono richiedere
   //from_config stabilisce se i parametri sono tutti contenuti in args (false) o se devono anche essere recuperati
   //dalla configurazione del motore Snmp locale (true)
   //restituisce la PDU di risposta o null se ci sono stati problemi
   */
    public SnmpPDU snmpGet(PlaceID placeID,String args[],String OIDs[],boolean from_config)
    {
       String host = conf.PlaceToHost(placeID);
       return(snmpGet(host,args,OIDs,from_config));
    }

   /** effettua una richiesta di 1 o piu' variabili Snmp su un insieme di agenti Snmp
   *   aprendo una sessione per ogni agente target e utilizzando richieste asincrone
   *   con meccanismo di callback
   *   places e' la lista dei place che ospitano gli agenti da interrogare
   *   ports e' la lista delle porte d'ascolto degli agenti Snmp
   *   args e' il vettore dei parametri: la posizione di ogni parametro e' definita dalle variabili
   *   static della classe Settings ed inizialmente conterra' solo i parametri opzionali eventuali da
   *   associare ad ogni richiesta,quindi: contextName,contextID,timeout,retries
   *   OIDs e' il vettore contenente tutti gli OID che si vogliono richiedere
   */
    public SnmpPDU[] snmpGet(String[] hosts,int[] ports,String args[],String[] OIDs)
   	{
        int NP = hosts.length;
		SnmpSession[] sessions = new SnmpSession[NP];
		int[] reqIDs = new int[NP];
   		
   		//creo i ricevitori asincroni da associare ad ogni sessione
   		ReceiverAtomicReq[] receivers = new ReceiverAtomicReq[NP];
   		for(int j=0; j<NP; j++)
   		{
   			sessions[j] = new SnmpSession(api);
   			receivers[j] = new ReceiverAtomicReq(debug,out);
   		}
   		Settings[] sets = prepareAsyncSessions(sessions,hosts,ports,args,receivers,false);

        //qui invio tutte le richieste asincrone sulle varie sessioni
        SnmpPDU pdu;
        for (int k=0; k<NP; k++)
        {
           	//faccio la richiesta asincrona lasciando al sistema la scelta del reqID
        	int id = 0;
        	pdu = snmpAsyncOp(sets[k],sessions[k],OIDs,GET,null,id);
        	//caso in cui la get non sia possibile
        	if (pdu == null)
        		reqIDs[k] = -1;
			else
        		reqIDs[k] = pdu.getReqid();
        }

        return (pickupAtomicAnswers(reqIDs,sessions,receivers));
	}
	
	
/*VERSIONE SEQUENZIALE	
public SnmpPDU[] snmpGet(String[] hosts,int[] ports,String args[],String[] OIDs)
    {
    	int NP = hosts.length;
   		Vector ris = new Vector();
   		
   		for (int j=0; j<NP; j++)
   		{
   			args[Settings.PORT] = (new Integer(ports[j])).toString();
   			ris.add(snmpGet(hosts[j],args,OIDs,true));
   		}
   		
   		SnmpPDU[] res = (SnmpPDU[])ris.toArray(new SnmpPDU[0]);
   		return res;
   	}
   		*/	
   		
   public SnmpPDU[] snmpGet(PlaceID[] places,int[] ports,String args[],String[] OIDs)
   {
	   	String[] hosts = new String[places.length];
	   	//recupero gli hosts corrispondenti ai places
		for (int j=0; j<places.length; j++)
            hosts[j] = conf.PlaceToHost(places[j]);

		return (snmpGet(hosts,ports,args,OIDs));
    }


	private SnmpPDU[] pickupAtomicAnswers(int[] reqIDs,SnmpSession[] sessions,ReceiverAtomicReq[] receivers)
	{
		int nreq = reqIDs.length;
     	SnmpPDU[] answers = new SnmpPDU[nreq];
     	boolean[] picked = new boolean[nreq];
     	for (int j=0; j<nreq; j++)
     	{
     		if (reqIDs[j] == -1)
     		{
     			nreq--;
     			picked[j] = true;
     		}
     		picked[j] = false;
     	}

     	while(nreq != 0)
        {
        	for(int k=0; k<sessions.length; k++)
        	{
				//se il ricevitore ha registrato di avere ricevuto la risposta
				//recupero la risposta, cancello la registrazione di un ricevitore per la ricezione
				//di risposte asincrone e chiudo la sessione
	        	if ( (picked[k] == false) && (receivers[k].hasDone()) )
	        	{
	        	   //System.out.println("PICK");
          			picked[k] = true;
        			answers[k] = receivers[k].getAnswer();
        			sessions[k].removeSnmpClient(receivers[k]);
        			sessions[k].close();
        			nreq--;
        		}
	        	//se e' scaduto il timeout sulla richiesta associata a questa sessione
	        	//devo solo registrare una risposta nulla e rimuovere il ricevente asincrono
	        	//oltre a chiudere la sessione stessa
	        	else if ( (picked[k] == false) && (sessions[k].checkTimeout(reqIDs[k])) )
	        	{
	        		answers[k] = null;
        			sessions[k].removeSnmpClient(receivers[k]);
        			sessions[k].close();
        			nreq--;
				}
        	}
        }
        return answers;
   }

   	private snmpTable[] pickupTableAnswers(SnmpSession[] sessions,ReceiverTableReq[] receivers)
	{
		//inizializzo il numero di risposte che mi aspetto
		int nansw = sessions.length;
     	snmpTable[] answers = new snmpTable[nansw];
     	boolean[] picked = new boolean[nansw];
     	for (int j=0; j<nansw; j++)
     		picked[j] = false;

        while(nansw != 0)
        {
        //System.out.println(nansw);
        	for(int k=0; k<sessions.length; k++)
        	{
        		//se il ricevitore asincrono ha registrato la fine del lavoro
        		//recupero il risultato
        		if ( (receivers[k].hasDone()) && (picked[k] == false) )
        		{
        //		System.out.println("PICK");
        			picked[k] = true;
        			answers[k] = receivers[k].getAnswer();
        			sessions[k].removeSnmpClient(receivers[k]);
        			sessions[k].close();
        			nansw--;
        		}

        		//se sulla sessione risultano richieste sulle quali e' scaduto il timeout
        		//non aspetto piu'
	        	else if (picked[k] == false)
	        		receivers[k].checkTimeout(sessions[k]);
        	}
        }

        return answers;
   }


   public SnmpPDU snmpGetNext(String host,String args[],String OIDs[],boolean from_config)
   {
        SnmpSession session = new SnmpSession(api);
        //se si richiede di ricorrere ai dati di configurazione si recuperano gli argomenti mancanti
    	//altrimenti e' gia' tutto in args
    	if (from_config)
    		if (prepareArguments(host,args) == false)
    			return null;

        //con tutti gli argomenti che ho preparo la sessione facendo gli opportuni settaggi
        Settings sets = prepareSession(host,args,from_config,session);

        SnmpPDU ris = snmpOp(sets,session,OIDs,GETNEXT,null);
        session.close();
        return ris;
   }

   public SnmpPDU snmpGetNext(PlaceID placeID,String args[],String OIDs[],boolean from_config)
    {
        String host = conf.PlaceToHost(placeID);
        return(snmpGetNext(host,args,OIDs,from_config));
    }

    public SnmpPDU[] snmpGetNext(String[] hosts,int[] ports,String args[],String[] OIDs)
   	{
   		int NP = hosts.length;

   		ReceiverAtomicReq[] receivers = new ReceiverAtomicReq[NP];
   		SnmpSession[] sessions = new SnmpSession[NP];
   		int[] reqIDs = new int[NP];
   		boolean from_config = false;

   		//qui preparo tutte le sessioni recuperando i parametri mancanit dalla configurazione
   		//e creo i ricevitori asincroni
   		for(int k=0;k<NP;k++)
   		{
        	//creo un gestore delle risposte asincrone indirizzate alla sessione appena creata
        	receivers[k] = new ReceiverAtomicReq(debug,out);
   			sessions[k] = new SnmpSession(api);
   		}
   		
   		Settings[] sets = prepareAsyncSessions(sessions,hosts,ports,args,receivers,false);
        //qui invio tutte le richieste asincrone sulle varie sessioni
        SnmpPDU pdu;
        for (int k=0; k<NP; k++)
        {
        	//faccio la richiesta asincrona lasciando al sistema la scelta del reqID
        	int id = 0;
        	pdu = snmpAsyncOp(sets[k],sessions[k],OIDs,GETNEXT,null,id);
        	//caso in cui la getnext non sia possibile
        	if (pdu == null)
        		reqIDs[k] = -1;
			else
	        	reqIDs[k] = pdu.getReqid();
        }

        return (pickupAtomicAnswers(reqIDs,sessions,receivers));

	}
	
	/*
    VERSIONE SEQUENZIALE	
	public SnmpPDU[] snmpGetNext(String[] hosts,int[] ports,String args[],String[] OIDs)
    {
    	int NP = hosts.length;
   		Vector ris = new Vector();
   		
   		for (int j=0; j<NP; j++)
   		{
   			args[Settings.PORT] = (new Integer(ports[j])).toString();
   			ris.add(snmpGetNext(hosts[j],args,OIDs,true));
   		}
   		
   		SnmpPDU[] res = (SnmpPDU[])ris.toArray(new SnmpPDU[0]);
   		return res;
	}*/

    public SnmpPDU[] snmpGetNext(PlaceID[] places,int[] ports,String args[],String[] OIDs)
    {
   		String[] hosts = new String[places.length];
	   	//recupero gli hosts corrispondenti ai places
		for (int j=0; j<places.length; j++)
            hosts[j] = conf.PlaceToHost(places[j]);

	    return (snmpGetNext(hosts,ports,args,OIDs));
   }

   /*esegue una richiesta GetBulk a partire dai parametri specificati in input:
   //ciascuno occupa una precisa posizione all'interno di args[]
   //apre, usa e chiude un'apposita sessione Snmp
   //args e' il vettore dei parametri: la posizione di ogni parametro e' definita dalle variabili
   //static della classe Settings
   //OIDs e' il vettore contenente tutti gli OID che si vogliono richiedere
   //non_rep indica per quanti OID si vuole un singola istanza
   //max_rep indica il numero massimo di istanze richieste per gli OID rimanenti
   //from_config stabilisce se i parametri sono tutti contenuti in args (false) o se devono anche essere recuperati
   //dalla configurazione del motore Snmp locale (true)
   //restituisce la PDU di risposta o null se ci sono stati problemi
   */

   public SnmpPDU snmpGetBulk(String host,String args[], String OIDs[], String non_rep, String max_rep, boolean from_config)
   {
        SnmpSession session = new SnmpSession(api);

        //se si richiede di ricorrere ai dati di configurazione si recuperano gli argomenti mancanti
    	//altrimenti e' gia' tutto in args
    	if (from_config)
    		if (prepareArguments(host,args) == false)
    			return null;

        //con tutti gli argomenti che ho preparo la sessione facendo gli opportuni settaggi
        Settings sets = prepareSession(host,args,from_config,session);

        String arr[] = new String[2];
        arr[0] = non_rep;
        arr[1] = max_rep;
        SnmpPDU ris = snmpOp(sets,session,OIDs,GETBULK,arr);
        session.close();
        return ris;
   }

   public SnmpPDU snmpGetBulk(PlaceID placeID,String args[],String OIDs[],String non_rep, String max_rep,boolean from_config)
    {
        String host = conf.PlaceToHost(placeID);
        return(snmpGetBulk(host,args,OIDs,non_rep,max_rep,from_config));

    }

   public SnmpPDU[] snmpGetBulk(String[] hosts,int[] ports,String args[],String[] OIDs,String non_rep, String max_rep)
   {
   		int NP = hosts.length;

        String arr[] = new String[2];
        arr[0] = non_rep;
        arr[1] = max_rep;

   		ReceiverAtomicReq[] receivers = new ReceiverAtomicReq[NP];
   		SnmpSession[] sessions = new SnmpSession[NP];
   		int[] reqIDs = new int[NP];
   		boolean from_config = false;

   		//qui preparo tutte le sessioni e creo i ricevitori asincroni
   		for(int k=0;k<NP;k++)
   		{
        	//creo un gestore delle risposte asincrone indirizzate alla sessione appena creata
        	receivers[k] = new ReceiverAtomicReq(debug,out);
   			sessions[k] = new SnmpSession(api);
   		}
   		
   		Settings[] sets = prepareAsyncSessions(sessions,hosts,ports,args,receivers,false);
   		
        //qui invio tutte le richieste asincrone sulle varie sessioni
        SnmpPDU pdu;
        for (int k=0; k<NP; k++)
        {
        	//faccio la richiesta asincrona lasciando al sistema la scelta del reqID
        	int id = 0;
        	pdu = snmpAsyncOp(sets[k],sessions[k],OIDs,GETBULK,arr,id);
        	//caso in cui la getbulk non sia possibile
        	if (pdu == null)
        		reqIDs[k] = -1;
        	else
        		reqIDs[k] = pdu.getReqid();
        }

        return (pickupAtomicAnswers(reqIDs,sessions,receivers));
   }
   
/*VERSIONE SEQUENZIALE	
   	public SnmpPDU[] snmpGetBulk(String[] hosts,int[] ports,String args[],String[] OIDs,String non_rep,String max_rep)
    {
    	int NP = hosts.length;
   		Vector ris = new Vector();
   		
   		for (int j=0; j<NP; j++)
   		{
   			args[Settings.PORT] = (new Integer(ports[j])).toString();
   			ris.add(snmpGetBulk(hosts[j],args,OIDs,non_rep,max_rep,true));
   		}
   		
   		SnmpPDU[] res = (SnmpPDU[])ris.toArray(new SnmpPDU[0]);
   		return res;
	}*/

    public SnmpPDU[] snmpGetBulk(PlaceID[] places,int[] ports,String args[],String[] OIDs,String non_rep,String max_rep)
   	{
   		String[] hosts = new String[places.length];
	   	//recupero gli hosts corrispondenti ai places
		for (int j=0; j<places.length; j++)
            hosts[j] = conf.PlaceToHost(places[j]);

		return (snmpGetBulk(hosts,ports,args,OIDs,non_rep,max_rep));
 	}

    /*percorre tutto l'albero di oggetti di un agente SNMP
    //args e' il vettore dei parametri: la posizione di ogni parametro e' definita dalle variabili
    //static della classe Settings
    //startOID e' l'oid da cui partire nella visita all'albero
    //from_config stabilisce se i parametri sono tutti contenuti in args (false) o se devono anche essere recuperati
    //dalla configurazione del motore Snmp locale (true)
    //restituisce la PDU di risposta o null se ci sono stati problemi
    */

	public void snmpWalk(String host,String args[], String startOID, boolean from_config)
	{
		String OID[] = new String[1];
		if (startOID == null)
			OID[0] = "1.1.0";
		else OID[0] = startOID;

        SnmpSession session = new SnmpSession(api);

        //se devo usare la configurazione recupero gli argomenti mancanti
        //altrimenti e' gia' tutto in args
        if (from_config)
    		if (prepareArguments(host,args) == false)
    			return ;
        //con tutti gli argomenti che ho preparo la sessione facendo gli opportuni settaggi
        Settings sets = prepareSession(host,args,from_config,session);

        SnmpPDU current = new SnmpPDU();
        boolean next = true;
        while (next)
        {
        	current = snmpOp(sets,session,OID,GETNEXT,null);
         	if (current == null)
        		next = false;
        	else
        	{
        	    if (current.getErrstat() != 0)
        	    {
        	    	//trace_err(current.getError());
        	    	next = false;
        	    }
        		else
        		{
            		trace(current.printVarBinds());
            		OID[0] = current.getObjectID(0).toString();
            	}
            }

        }

        return;
    }

   public void snmpWalk(PlaceID placeID,String args[], String startOID, boolean from_config)
   {
       String host = conf.PlaceToHost(placeID);

        //se devo usare la configurazione recupero gli argomenti mancanti
        //altrimenti e' gia' tutto in args
        if (from_config)
    		if (prepareArguments(host,args) == false)
    			return ;

        snmpWalk(host,args,startOID,from_config);
   }

   //Operazioni Snmp:
   //quasi tutti i parametri sono stati settati sulla sessione questo permette di non ripetere
   //i settaggi sulla PDU (il default e' rappr. dai valori sulla sessione)

   //questo metodo permette di fare richieste Snmp elementari sulla sessione passata come parametro
   //le operazioni per ora previste sono GET/GETNEXT/GETBULK
   //i parametri utilizzati per inviare la PDU sono quelli della sessione
   //session e' la sessione SNMP gia' aperta con gli opportuni parametri
   //sets e' un istanza di Settings che contiene tutte le informazioni sui settaggi associati
   //alla richiesta
   //OIDs e' la lista di OID desiderati
   //opcode mi permette di distinguere le operazioni, sono valori possibili:GET GETNEXT GETBULK
   //argsOp e' un array di argomenti per l'operazione richiesta; ad esempio con getBulk si trattera' di
   //non repetitions e maxrepetitions
   //restituisce la PDU di risposta oppure null se ci sono stati problemi
   SnmpPDU snmpOp(Settings sets,SnmpSession session,String OIDs[],int opcode,String argsOp[])
   {
        if (OIDs == null)
        {
            trace_err("At least one OID !!");
            return null;
        }
        //creo una nuova PDU che dovro' poi spedire
        SnmpPDU pdu = new SnmpPDU();
        //setto il tipo di operazione nella pdu
        switch (opcode)
        {
            case GET: pdu.setCommand(api.GET_REQ_MSG);
                      break;
            case GETNEXT: pdu.setCommand(api.GETNEXT_REQ_MSG);
                          break;
            case GETBULK: if (session.getVersion() == SnmpAPI.SNMP_VERSION_1)
                          {
                            trace_err("Snmpv1 doesn't support GetBulk");
                            return null;
                          }
                          pdu.setCommand(api.GETBULK_REQ_MSG);
                          pdu.setErrstat(Integer.parseInt(argsOp[0]));
                          pdu.setErrindex(Integer.parseInt(argsOp[1]));
                          break;
            default: trace_err("Operation not supported. Available: GET/GETNEXT/GETBULK");
        }

        //introduco tutti gli OID richiesti nella PDU
        for (int j=0; j < OIDs.length; j++)
        {
            SnmpOID oid = new SnmpOID(OIDs[j]);
            if (oid.toValue() == null)
            {
                trace_err("Invalid OID argument: "+OIDs[j]);
                return null;
            }
            else pdu.addNull(oid);
        }
		
		//la parte successiva viene eseguita solo se e' la prima volta che si usa la sessione
		//con nuovi settaggi
		pdu.setUserName(sets.userName.getBytes());
		if (session.getVersion() == SnmpAPI.SNMP_VERSION_3)
		{
			if (sets.config == false)
			USMUtils.init_v3_params(sets.userName,sets.authProto,sets.authPwd,sets.privPwd,session.getPeername(),session.getRemotePort(),session);
			pdu.setContextName(sets.contextName.getBytes());
			pdu.setContextID(sets.contextID.getBytes());
		}
				
        try  {  if (debug) 
        		api.setDebug(true);
				pdu = session.syncSend(pdu); }
        catch (SnmpException e)
        {
            trace_err("Error on sending PDU "+e.getMessage());
            return null;
        }

        return(pdu);
   }
   
   
   SnmpPDU snmpOp(SnmpSession session,String OIDs[],int opcode,String argsOp[])
   {
        if (OIDs == null)
        {
            trace_err("At least one OID !!");
            return null;
        }
        //creo una nuova PDU che dovro' poi spedire
        SnmpPDU pdu = new SnmpPDU();
        //setto il tipo di operazione nella pdu
        switch (opcode)
        {
            case GET: pdu.setCommand(api.GET_REQ_MSG);
                      break;
            case GETNEXT: pdu.setCommand(api.GETNEXT_REQ_MSG);
                          break;
            case GETBULK: if (session.getVersion() == SnmpAPI.SNMP_VERSION_1)
                          {
                            trace_err("Snmpv1 doesn't support GetBulk");
                            return null;
                          }
                          pdu.setCommand(api.GETBULK_REQ_MSG);
                          pdu.setErrstat(Integer.parseInt(argsOp[0]));
                          pdu.setErrindex(Integer.parseInt(argsOp[1]));
                          break;
            default: trace_err("Operation not supported. Available: GET/GETNEXT/GETBULK");
        }

        //introduco tutti gli OID richiesti nella PDU
        for (int j=0; j < OIDs.length; j++)
        {
            SnmpOID oid = new SnmpOID(OIDs[j]);
            if (oid.toValue() == null)
            {
                trace_err("Invalid OID argument: "+OIDs[j]);
                return null;
            }
            else pdu.addNull(oid);
        }

        try  {  if (debug) 
        		api.setDebug(true);
				pdu = session.syncSend(pdu); }
        catch (SnmpException e)
        {
            trace_err("Error on sending PDU "+e.getMessage());
            return null;
        }

        return(pdu);
   }

//DA MODIFICARE PER LIMITARE LA REPLICAZ. DI CODICE
   SnmpPDU snmpAsyncOp(Settings sets,SnmpSession session,String OIDs[],int opcode,String argsOp[],int reqID)
   {
        if (OIDs == null)
        {
            trace_err("At least one OID !!");
            return null;
        }
        //creo una nuova PDU che dovro' poi spedire
        SnmpPDU pdu = new SnmpPDU();
        //setto il tipo di operazione nella pdu
        switch (opcode)
        {
            case GET: pdu.setCommand(api.GET_REQ_MSG);
                      break;
            case GETNEXT: pdu.setCommand(api.GETNEXT_REQ_MSG);
                          break;
            case GETBULK: if (session.getVersion() == SnmpAPI.SNMP_VERSION_1)
                          {
                            trace_err("Snmpv1 doesn't support GetBulk");
                            return null;
                          }
                          pdu.setCommand(api.GETBULK_REQ_MSG);
                          pdu.setErrstat(Integer.parseInt(argsOp[0]));
                          pdu.setErrindex(Integer.parseInt(argsOp[1]));
                          break;
            default: trace_err("Operation not supported. Available: GET/GETNEXT/GETBULK");
        }

        //introduco tutti gli OID richiesti nella PDU
        for (int j=0; j < OIDs.length; j++)
        {
            SnmpOID oid = new SnmpOID(OIDs[j]);
            if (oid.toValue() == null)
            {
                trace_err("Invalid OID argument: "+OIDs[j]);
                return null;
            }
            else pdu.addNull(oid);
        }
        
        //la parte successiva viene eseguita solo se i settaggi sono diversi da null
        //cioe' se e' la prima volta che si usa questa sessione con nuovi settaggi
		pdu.setUserName(sets.userName.getBytes());
		if (session.getVersion() == SnmpAPI.SNMP_VERSION_3)
		{
			if (sets.config == false)
			USMUtils.init_v3_params(sets.userName,sets.authProto,sets.authPwd,sets.privPwd,session.getPeername(),session.getRemotePort(),session);
			pdu.setContextName(sets.contextName.getBytes());
			pdu.setContextID(sets.contextID.getBytes());
		}
				
        //se e' stato specificato un reqID != 0 significa che se ne vuole attribuire uno proprio
        //se e' = 0 lo assegna il sistema e se ne conosce il valore al ritorno dalla send
		if (reqID != 0)
			pdu.setReqid(reqID);
        try
        {  	session.send(pdu);   }
        catch (SnmpException e)
        {
            trace_err("Error on sending PDU "+e.getMessage());
            return null;
        }

        return(pdu);
   }
   
    //recupera l'intera tabella associata all'oid passato come argomento
    //args e' il vettore dei parametri: la posizione di ogni parametro e' definita
    //dalle variabili static della classe Settings
    //N.B. apre,usa,chiude una sessione Snmp apposita
    //oid_tab e' l'oid, sotto forma di stringa, che identifica la tabella
    //from_config stabilisce se i parametri sono tutti contenuti in args (false) o se devono anche essere recuperati
    //dalla configurazione del motore Snmp locale (true)
    //restituisce la tabella delle SnmpVar recuperate per ogni elemento della tabella
    public snmpTable getTable(String host,String args[], String oid_table, boolean from_config)
    {
        SnmpSession session = new SnmpSession(api);

        //se si richiede di ricorrere ai dati di configurazione si recuperano gli argomenti mancanti
    	//altrimenti e' gia' tutto in args
    	if (from_config)
    		if (prepareArguments(host,args) == false)
    			return null;

        //con tutti gli argomenti che ho preparo la sessione facendo gli opportuni settaggi
        Settings sets = prepareSession(host,args,from_config,session);

        snmpTable table = getTable(sets,session,oid_table);
        session.close();
        return table;
    }

    public snmpTable getTable(PlaceID placeID,String args[], String oid_table, boolean from_config)
    {
       String host = conf.PlaceToHost(placeID);

       return(getTable(host,args,oid_table,from_config));
    }

    //recupera l'intera tabella associata all'oid passato come argomento
    //session e' la sessione Snmp che si suppone gia' aperta e settata con gli opportuni parametri
    //sets_in_request e' l'insieme dei settings associati alla richiesta
    //oid_tab e' l'oid, sotto forma di stringa, che identifica la tabella
    //restituisce la tabella delle SnmpVar recuperate per ogni elemento della tabella
    public snmpTable getTable(Settings sets,SnmpSession session,String oid_table)
    {
    	String remaddr = session.getPeername();
    	int remport = session.getRemotePort();
        MibNode nodetab = conf.getloadMibTable(remaddr,remport,oid_table);
        if (nodetab == null)
        	return null;
        System.out.println(nodetab.getLabel());
        String OIDs[] = conf.getColTable(nodetab);
        if (OIDs == null)
        	return null;

        boolean next = true;
        int nrow = 0;
        SnmpPDU ris = new SnmpPDU();
        snmpTable table = new snmpTable(oid_table,OIDs.length,debug,out);
        while (next)
        {
            ris = snmpOp(sets,session,OIDs,GETNEXT,null);
            next = (check(ris,oid_table));
            if (next)
            {
                table.addRow(ris);
                nrow++;
                //recupero gli OID della risposta per utilizzarli nella prossima richiesta
                for(int i=0; i<OIDs.length; i++)
                    OIDs[i] = ris.getObjectID(i).toString();
            }
            else //errore o risposta non piu' relativa alla tabella
            {
                next = false;
                break;
            }
        }
        System.out.println("RIGHE: "+nrow);
        return table;
       }

	public snmpTable[] getTable(String[] hosts,int[] ports,String args[],String OIDtab)
    {
    	int NP = hosts.length;
   		ReceiverTableReq[] receivers = new ReceiverTableReq[NP];
   		SnmpSession[] sessions = new SnmpSession[NP];
   		int[] reqIDs = new int[NP];

       	//suppongo che tutti gli agenti specificati supportino un Mib contenente la
       	//tabella ed utilizzo arbitrariamente il primo per recuperare il nome delle colonne
		MibNode nodetab = conf.getloadMibTable(hosts[0],ports[0],OIDtab);
		if (nodetab == null)
			return null;
		String[] cols = conf.getColTable(nodetab);
        if (cols == null)
        	return null;
   		for(int k=0;k<NP;k++)
   		{
   			sessions[k] = new SnmpSession(api);
   			//creo la tabella
        	snmpTable tab = new snmpTable(OIDtab,cols.length,debug,out);
        	//creo un gestore delle risposte asincrone indirizzate alla sessione
        	receivers[k] = new ReceiverTableReq(this,tab,k+1,debug,out);
        }
        Settings[] sets = prepareAsyncSessions(sessions,hosts,ports,args,receivers,true);

        //qui invio tutte le richieste asincrone sulle varie sessioni della prima riga,
        //le successive verranno fatte nella callback del Receiver
        SnmpPDU pdu;
        for (int k=0; k<NP; k++)
        {
        	receivers[k].initparams(sets[k]);
			pdu = snmpAsyncOp(sets[k],sessions[k],cols,GETNEXT,null,k+1);

       		reqIDs[k] = pdu.getReqid();
        }

        return (pickupTableAnswers(sessions,receivers));
    }
    
/*VERSIONE SEQUENZIALE	
	public snmpTable[] getTable(String[] hosts,int[] ports,String args[],String OIDtab)
    {
    	int NP = hosts.length;
   		Vector ris = new Vector();
   		
   		for (int j=0; j<NP; j++)
   		{
   			args[Settings.PORT] = (new Integer(ports[j])).toString();
   			ris.add(getTable(hosts[j],args,OIDtab,true));
   		}
   		
   		snmpTable[] res = (snmpTable[])ris.toArray(new snmpTable[0]);
   		return res;
   	}*/
   			
    public snmpTable[] getTable(PlaceID[] places,int[] ports,String args[],String OIDtab)
   	{
   		String[] hosts = new String[places.length];
	   	//recupero gli hosts corrispondenti ai places
		for (int j=0; j<places.length; j++)
            hosts[j] = conf.PlaceToHost(places[j]);

		return (getTable(hosts,ports,args,OIDtab));
	}

    //per ora considero la possibilita' di specificare > < = un valore
    /**recupera le righe della tabella, associata all'oid passato come argomento, che verificano
    //una condizione
    //args e' il vettore dei parametri: la posizione di ogni parametro e' definita
    //dalle variabili static della classe Settings
    //N.B. apre,usa,chiude una sessione Snmp apposita
    //oid_tab e' l'oid, sotto forma di stringa, che identifica la tabella
    //oid_col e' l'oid della colonna sulla quale deve essere verificata la condizione
    //cond e' la condizione (">"\"<"\"=")
    //rif e' il valore di riferimento per la colonna
    //from_config stabilisce se i parametri sono tutti contenuti in args (false) o se devono anche essere recuperati
    //dalla configurazione del motore Snmp locale (true)
    //restituisce la tabella delle SnmpVar recuperate per le righe che verificano la condizione
    */
    public snmpTable FilterTable(String host,String args[], String oid_tab,String rif1,String cond1,String oid_col,String cond2,String rif2,boolean from_config)
    {
        if (oid_col.startsWith(oid_tab) == false)
        {
            trace_err("Specified column doesn't belong to specified table");
            return null;
        }
        SnmpSession session = new SnmpSession(api);
        //se si richiede di ricorrere ai dati di configurazione si recuperano gli argomenti mancanti
    	//altrimenti e' gia' tutto in args
    	if (from_config)
    		if (prepareArguments(host,args) == false)
    			return null;

        //con tutti gli argomenti che ho preparo la sessione facendo gli opportuni settaggi
        Settings sets = prepareSession(host,args,from_config,session);
        snmpTable table = FilterTable(sets,session,oid_tab,rif1,cond1,oid_col,cond2,rif2);
        session.close();
        return table;

    }

    public snmpTable FilterTable(PlaceID place,String args[], String oid_tab,String rif1,String cond1,String oid_col,String cond2,String rif2,boolean from_config)
    {
       String host = conf.PlaceToHost(place);

       return(FilterTable(host,args,oid_tab,rif1,cond1,oid_col,cond2,rif2,from_config));
    }

    /**recupera le righe della tabella, associata all'oid passato come argomento, che verificano
    //una condizione
    //session e' la sessione Snmp che si suppone gia' aperta e settata con gli opportuni parametri
    //sets_in_request e' l'insieme dei settings associati alla richiesta
    //oid_tab e' l'oid, sotto forma di stringa, che identifica la tabella
    //oid_col e' l'oid della colonna sulla quale deve essere verificata la condizione
    //cond e' la condizione (">"\"<"\"=")
    //rif e' il valore di riferimento per la colonna
    //restituisce la tabella delle SnmpVar recuperate per le righe che verificano la condizione
    */
    public snmpTable FilterTable(Settings sets,SnmpSession session,String oid_tab,String rif1,String cond1,String oid_col,String cond2,String rif2)
    {
        if (oid_col.startsWith(oid_tab) == false)
        {
            trace_err("Specified column doesn't belong to specified table");
            return null;
        }

        //risalgo al numero di colonna costruendo la stringa che rimane se si toglie all'oid della colonna
        //quello della tabella ed il "." es "1.2.3.1" - "1.2.3." --> "1"
        int index = oid_col.lastIndexOf('.');
        String numcol = oid_col.substring(++index);
        int Ncol = Integer.parseInt(numcol);
        String remaddr = session.getPeername();
        int remport = session.getRemotePort();
        MibNode nodetab = conf.getloadMibTable(remaddr,remport,oid_tab);
        if (nodetab == null)
        	return null;
        System.out.println("Target Table: "+nodetab.getLabel());
        String OIDs[] = conf.getColTable(nodetab);
        if (OIDs == null)
        	return null;

        boolean next = true;
        int nrow = 0;
        SnmpPDU ris = new SnmpPDU();
        snmpTable table = new snmpTable(oid_tab,OIDs.length,debug,out);
        while (next)
        {
            ris = snmpOp(sets,session,OIDs,GETNEXT,null);
            next = (check(ris,oid_tab));
            if (next)
            {
                if (verifyRange(ris.getVariable(Ncol-1),rif1,cond1,cond2,rif2))
                {
                    table.addRow(ris);
                    nrow++;
                }
                //recupero gli OID della risposta per utilizzarli nella prossima richiesta
                for(int i=0; i<OIDs.length; i++)
                    OIDs[i] = ris.getObjectID(i).toString();
            }
            else //errore o risposta non piu' relativa alla tabella
            {
                next = false;
                break;
            }
        }
        System.out.println(nrow+" Rows Matching Condition: "+rif1+cond1+oid_col+cond2+rif2);
        return table;
    }

    public snmpTable[] FilterTable(String[] hosts,int[] ports,String args[],String oid_tab,String rif1,String cond1,String oid_col,String cond2,String rif2)
    {
    	int NP = hosts.length;
   		Vector ris = new Vector();
   		
   		for (int j=0; j<NP; j++)
   		{
   			args[Settings.PORT] = (new Integer(ports[j])).toString();
   			ris.add(FilterTable(hosts[j],args,oid_tab,rif1,cond1,oid_col,cond2,rif2,true));
   		}
   		
   		snmpTable[] res = (snmpTable[])ris.toArray(new snmpTable[0]);
   		return res;
   	}
   	
   public snmpTable[] FilterTable(PlaceID[] places,int[] ports,String args[],String oid_tab,String rif1,String cond1,String oid_col,String cond2,String rif2)
   	{
   		String[] hosts = new String[places.length];
	   	//recupero gli hosts corrispondenti ai places
		for (int j=0; j<places.length; j++)
            hosts[j] = conf.PlaceToHost(places[j]);

		return (FilterTable(hosts,ports,args,oid_tab,rif1,cond1,oid_col,cond2,rif2));
	}

	/**recupera informazioni relative a 1 o piu' variabili Snmp con riferimento
	*  a piu' places target o, nel caso piu' generale, ad un dominio
	*  di ogni variabile Snmp si riportano il valore massimo, minimo, medio rilevati sull'insieme di
	*  targets, tenendo anche traccia dell'agente Snmp (o "degli" in caso di parita') che hanno restituito tali valori
	*  restituisce una struttura InfoDomain contenente tutte queste informazioni
	*  places sono i place target
	*  ports sono le porte d'ascolto degli agenti Snmp, sui place specificati come target
	*  args sono i parametri comuni (contextName,contextID,timeout,retries) da utilizzare nelle richieste (gli altri
	*  si recuperano da configurazione)
	*  OIDs sono gli identificatori delle variabili Snmp di interesse, devono essere variabili numeriche
	*/
    public InfoDomain getInfoOnDomain(PlaceID[] places,int[] ports,String args[],String[] OIDs)
	{
		//QUI POTREI INSERIRE UN CONTROLLO SUGLI OID PER ACCERTARMI CHE SIANO TUTTE VAR NUMERICHE

		SnmpPDU[] ris = snmpGet(places,ports,args,OIDs);
		InfoDomain idom = new InfoDomain(OIDs.length);
		//processo l'array di SnmpPDU di risposta per determinare valori min,max,medi
		for (int j=0; j<ris.length; j++)
		{
			SnmpPDU pdu = ris[j];
			Vector varbs = pdu.getVariableBindings();
			InetAddress host = conf.getInfo(places[j]).host;
			IpAddress ip = new IpAddress(host,ports[j]);
			int k=0;
			//per ogni valore che mi ritorna dalla risposta di un agente effettuo l'eventuale aggiornamento
			//di valore max/min e aggiungo alla somma di tutti i valori della variabile reperiti dagli
			//agenti fino ad ora interrogati quello corrente
			for(Enumeration e=varbs.elements(); e.hasMoreElements();k++)
			{
				SnmpVar v = ((SnmpVarBind)e.nextElement()).getVariable();
			    idom.setCurMinMax(v,k,">",ip);
				idom.setCurMinMax(v,k,"<",ip);
				idom.setCurSum(v,k);
			}
		}

		//fatta la somma dei valori di tutti gli agenti interrogati posso calcolare la media
		idom.setAve();

		return idom;
	}

	/**recupera informazioni relative a 1 o piu' variabili Snmp con riferimento
	*  a piu' agenti Snmp target o
	*  di ogni variabile Snmp si riportano il valore massimo, minimo, medio rilevati sull'insieme di
	*  targets, tenendo anche traccia dell'agente Snmp (o "degli" in caso di parita') che hanno restituito tali valori
	*  restituisce una struttura InfoDomain contenente tutte queste informazioni
	*  hosts sono gli hosts su cui si trovano gli agenti Snmp target
	*  ports sono le porte d'ascolto degli agenti Snmp, sui place specificati come target
	*  args sono i parametri comuni (contextName,contextID,timeout,retries) da utilizzare nelle richieste (gli altri
	*  si recuperano da configurazione)
	*  OIDs sono gli identificatori delle variabili Snmp di interesse, devono essere variabili numeriche
	*/
    public InfoDomain getInfoOnDomain(String[] hosts,int[] ports,String args[],String[] OIDs)
	{
		//QUI POTREI INSERIRE UN CONTROLLO SUGLI OID PER ACCERTARMI CHE SIANO TUTTE VAR NUMERICHE

		SnmpPDU[] ris = snmpGet(hosts,ports,args,OIDs);
		InfoDomain idom = new InfoDomain(OIDs.length);
		//processo l'array di SnmpPDU di risposta per determinare valori min,max,medi
		for (int j=0; j<ris.length; j++)
		{
			SnmpPDU pdu = ris[j];
			Vector varbs = pdu.getVariableBindings();
			IpAddress ip = new IpAddress(hosts[j],ports[j]);
			int k=0;
			//per ogni valore che mi ritorna dalla risposta di un agente effettuo l'eventuale aggiornamento
			//di valore max/min e aggiungo alla somma di tutti i valori della variabile reperiti dagli
			//agenti fino ad ora interrogati quello corrente
			for(Enumeration e=varbs.elements(); e.hasMoreElements();k++)
			{
				SnmpVar v = ((SnmpVarBind)e.nextElement()).getVariable();
			    idom.setCurMinMax(v,k,">",ip);
				idom.setCurMinMax(v,k,"<",ip);
				idom.setCurSum(v,k);
			}
		}

		//fatta la somma dei valori di tutti gli agenti interrogati posso calcolare la media
		idom.setAve();

		return idom;
	}

    /** permette di eseguire il polling, cioe' di richiedere per un certa finestra
    //temporale la ripetizione ad intervalli regolari di una operazione Snmp
    //elementare o derivata: GET/GETNEXT/GETBULK su uno o piu' OID,recupero di un'intera tabella
    //o filtraggio di una tabella tramite condizione su una colonna
    //args e' il vettore dei parametri: la posizione di ogni parametro e' definita dalle variabili
    //static della classe Settings
    //extra e' un array di argomenti specifici per l'operazione richiesta:
    //ad esempio con getBulk si avranno non_repetitions e max_repetitions; con filtert si avra'
    //il nome della colonna di riferimento, il valore di riferimento, la condizione
    //OIDs e' la lista di OID coinvolti
    //op_type mi permette di distinguere le operazioni, sono valori possibili:
    //GET GETNEXT GETBULK GETABLE FILTERT
    //interval e' la finestra temporale (in millisec) di osservazione
    //delay e' l'intervallo di polling (in millisec)
    //
    */
    public Vector snmpPoller(String host,String args[],String extra[],String OIDs[],int op_type,int interval,int delay, boolean from_config)
    {
        SnmpSession session = new SnmpSession(api);

        //se si richiede di ricorrere ai dati di configurazione si recuperano gli argomenti mancanti
    	//altrimenti e' gia' tutto in args
    	if (from_config)
    		if (prepareArguments(host,args) == false)
    			return null;

        //con tutti gli argomenti che ho preparo la sessione facendo gli opportuni settaggi
        Settings sets = prepareSession(host,args,from_config,session);

        Pollerb poll=new Pollerb(this,sets,session,OIDs,extra,op_type,interval,delay, new sharedOut(out));
        Vector ris = poll.run();
        session.close();
        return ris;
    }

    public Vector snmpPoller(PlaceID place,String args[],String extra[],String OIDs[],int op_type,int interval,int delay, boolean from_config)
    {
       String host = conf.PlaceToHost(place);

       return (snmpPoller(host,args,extra,OIDs,op_type,interval,delay,from_config));
    }


	//il polling su piu' nodi restituisce un vettore per ogni target contenente tutti i campioni
    public Vector snmpPoller(String hosts[],int[] ports,String args[],String extra[],String OIDs[],int op_type,int interval,int delay)
    {
		int N = hosts.length;
   		Vector ris = new Vector();
		for (int j=0; j<N; j++)
		{
			args[Settings.PORT] = String.valueOf(ports[j]);
			ris.add(snmpPoller(hosts[j],args,extra,OIDs,op_type,interval,delay,true));
		}
		return ris;
    }

    public Vector snmpPoller(PlaceID[] places,int[] ports,String args[],String extra[],String OIDs[],int op_type,int interval,int delay)
    {
   		String[] hosts = new String[places.length];
	   	//recupero gli hosts corrispondenti ai places
		for (int j=0; j<places.length; j++)
            hosts[j] = conf.PlaceToHost(places[j]);

		return (snmpPoller(hosts,ports,args,extra,OIDs,op_type,interval,delay));
    }

    //verifica se la PDU e' OK
    //pdu e' la pdu ottenuta in risposta ad una richiesta SNMP
    //restituisce true se la pdu e' OK
    //false se e' nulla o se contiene indicazioni di errore
    boolean check(SnmpPDU pdu,String tab)
    {
        //verifico se la pdu e' nulla (time out)
        if (pdu == null)
        {
            trace_err("NULL PDU in response");
            return false;
        }

        //verifico se la pdu contiene errori
        if (pdu.getErrstat() > 0)
        {
            trace_err("Error occurred in PDU "+pdu.getError());
            pdu.fix();
            return false;
        }

        //verifico se nella pdu di risposta sono contenuti elementi appartenenti alla tabella
        int nvar = (pdu.getVariableBindings()).size();
        int j = 0;
        boolean next = true;
        while ((next)&&(j<nvar))
    	    next = pdu.getObjectID(j++).toString().startsWith(tab+".");

        return next;
    }

	public boolean checkPDU(SnmpPDU pdu)
	{
		//verifico se la pdu e' nulla (time out)
        if (pdu == null)
        {
            trace_err("NULL PDU in response");
            return false;
        }

        //verifico se la pdu contiene errori
        if (pdu.getErrstat() > 0)
        {
            trace_err("Error occurred in PDU "+pdu.getError());
            pdu.fix();
       		return false;
        }

        return true;
    }

    /** verifica la condizione <inf> <condinf> <elem> <condsup> <sup>
    //elem e' la SnmpVar sulla quale si deve fare la verifica
    //condinf e' la condizione inferiore che puo' assumere i valori "<" "<="
    //condsup e' la condizione superiore
    //inf e' il valore di riferimento inferiore
    //sup e' il valore di riferimento superiore
    */
    boolean verifyRange(SnmpVar elem,String rif_inf,String cond_inf, String cond_sup,String rif_sup)
    {
        byte eltype = elem.getType();
        int ret1;
        int ret2;
	    //col < rif o col <= rif
        if (rif_inf.equals("ANY"))
            return(verifyCond(elem,cond_sup,rif_sup));
        //rif < col o rif <= col
        if (rif_sup.equals("ANY"))
        {
        	if (cond_inf.endsWith("="))
        		cond_inf = ">=";
			else
				cond_inf = ">";
            return(verifyCond(elem,cond_inf,rif_inf));
        }
        else
        {
            if (eltype == SnmpAPI.INTEGER)
            {
                Integer cur = (Integer)elem.toValue();
                Integer target1 = new Integer(rif_inf);
                Integer target2 = new Integer(rif_sup);
                ret1 = target1.compareTo(cur);
                ret2 = cur.compareTo(target2);

            }
            else if ((eltype == SnmpAPI.COUNTER) || (eltype == SnmpAPI.COUNTER64) ||
                (eltype == SnmpAPI.GAUGE) || (eltype == SnmpAPI.TIMETICKS))
            {
                Long cur = (Long)elem.toValue();
                Long target1 = new Long(rif_inf);
                Long target2 = new Long(rif_sup);
                ret1 = target1.compareTo(cur);
                ret2 = cur.compareTo(target2);
            }
            else if ((eltype == SnmpAPI.BITSTRING) || (eltype == SnmpAPI.IPADDRESS) ||
                (eltype == SnmpAPI.NETWORKADDRESS) || (eltype == SnmpAPI.NSAP) ||
                (eltype == SnmpAPI.OPAQUE) || (eltype == SnmpAPI.STRING) ||
                (eltype == SnmpAPI.OBJID))
            {
                String cur = (String) elem.toString();
                System.out.println(cur);
                String target1 = new String(rif_inf);
                ret1 = target1.compareTo(cur);
                String target2 = new String(rif_sup);
                ret2 = cur.compareTo(target2);
            }

            else
                return false;

            if ( (verify(ret1,cond_inf)) && (verify(ret2,cond_sup)) )
                return true;

            return false;
        }
    }

    /** verifica la condizione <elem> <cond> <rif>
    //elem e' la SnmpVar sulla quale si deve fare la verifica
    //cond e' la condizione che puo' assumere i valori "<" "<=" ">" ">="
    //rif e' il valore di riferimento
    */
    boolean verifyCond(SnmpVar elem,String cond,String rif)
    {
        byte eltype = elem.getType();
        int ret;

        if (rif.equals("ANY"))
            return true;
        if (eltype == SnmpAPI.INTEGER)
        {
            Integer cur = (Integer)elem.toValue();
            Integer target = new Integer(rif);
            ret = cur.compareTo(target);
        }
        else if ((eltype == SnmpAPI.COUNTER) || (eltype == SnmpAPI.COUNTER64) ||
            (eltype == SnmpAPI.GAUGE) || (eltype == SnmpAPI.TIMETICKS))
        {
            Long cur = (Long)elem.toValue();
            Long target = new Long(rif);
            ret = cur.compareTo(target);
        }
        else if ((eltype == SnmpAPI.BITSTRING) || (eltype == SnmpAPI.IPADDRESS) ||
            (eltype == SnmpAPI.NETWORKADDRESS) || (eltype == SnmpAPI.NSAP) ||
            (eltype == SnmpAPI.OPAQUE) || (eltype == SnmpAPI.STRING) ||
            (eltype == SnmpAPI.OBJID))
        {
            String cur = (String) elem.toString();
            System.out.println(cur);
            String target = new String(rif);
            ret = cur.compareTo(target);
        }

        else
            return false;

        return (verify(ret,cond));

    }

    boolean verify(int ret,String cond)
    {
        if ( (ret == 0) && ( (cond.equals("=")) || (cond.equals(">=")) || (cond.equals("<=")) ) )
            return true;
        if ((ret < 0) && (cond.startsWith("<")))
            return true;
        if ((ret > 0) && (cond.startsWith(">")))
            return true;

        return false;
    }



    /*esegue la valutazione di una health function specificata dall'utente a partire dai parametri specificati in
   //input: ciascuno occupa una precisa posizione all'interno di args[]
   //host e' l'indirizzo presso cui si trova l'agente Snmp da interrogare
   //args e' il vettore dei parametri: la posizione di ogni parametro e' definita dalle variabili
   //static della classe Settings
   //HF e' l'health function sotto forma di stringa espressa in funzione di variabili Snmp in forma numerica e/o simbolica
   //from_config stabilisce se i parametri sono tutti contenuti in args (false) o se devono anche essere recuperati
   //dalla configurazione del motore Snmp locale (true)
   //restituisce l'Object (Integer/Long) di risposta o null se ci sono stati problemi
   */
  	public Object snmpHF(String host,String args[],String HF,boolean from_config)
  	{
        SnmpSession session = new SnmpSession(api);

        //se si richiede di ricorrere ai dati di configurazione si recuperano gli argomenti mancanti
    	//altrimenti e' gia' tutto in args
    	if (from_config)
    		if (prepareArguments(host,args) == false)
    			return null;

        //con tutti gli argomenti che ho preparo la sessione facendo gli opportuni settaggi
        Settings sets = prepareSession(host,args,from_config,session);
        
        return(snmpHF(sets,session,HF));

  	}

  	public Object snmpHF(PlaceID place,String args[],String HF,boolean from_config)
  	{
       String host = conf.PlaceToHost(place);

       return(snmpHF(host,args,HF,from_config));
    }

	//idem ma utilizzando una sessione di interazione gia' aperta e pronta (opportuni parametri gia' settati)
  	public Object snmpHF(Settings sets,SnmpSession session,String HF)
  	{
  		//ricavo dalla stringa che rappresenta l'health function le varibili Snmp coinvolte e le traduco nei corrispondenti OID
  		//numerici
  		//N.B. prevedo che cio' che mi viene restituito sia una lista degli OID non ripetuti (vengono eliminati gli eventuali duplicati)
  		String[] res = parseHF(HF);
  		if (res == null)
  			return null;

  		 HF = res[res.length-1];
  		 String oids[] = new String[res.length-1];
  		 for (int i=0; i<res.length-1; i++)
  		 	oids[i] = res[i];

  		//richiedo il reperimento del valore associato alle variabili tramite una snmpGet normale
  		SnmpPDU pdu = snmpOp(sets,session,oids,GETNEXT,null);
  		//ora recupero dalla pdu di risposta il valore di ogni var e lo associo all'oid corrispondente in una hashtable cosi' da
  		//poterli poi recuperare  facilmente al momento della valutazione
  		Hashtable values = new Hashtable();
  		int j =0;
  		for (Enumeration e = pdu.getVariableBindings().elements(); e.hasMoreElements(); )
  		{
  			SnmpVarBind varb = (SnmpVarBind) e.nextElement();
  			Object val = varb.getVariable().toValue();
  	System.out.println(oids[j]+" : "+val);
  			try
  			{  values.put(oids[j++],val);  }
  			catch (NullPointerException ex) { trace_err("Value of variable "+oids[j]+" not available");  return null;}
  		}

  		//chiamo la funzione di valutazione della HF passando come primo operando null, come numero di parentesi aperte 0 e
  		//trasmettendo la conoscenza delle associazioni variabile-valore
  		return (evalHF(new ret(HF), values).ris);
  	}



	
	/**calcola il valore di una HF per ogni host specificato, utilizzando i parametri di richiesta specificati in args
	*  restituisce un array di Object, che saranno Integer o Long a seconda del tipo delle variabili coinvolte nella HF 	
	*/
	public Object[] snmpHF(String[] hosts,int[] ports,String args[],String HF)
   	{
   	  	//ricavo dalla stringa che rappresenta l'health function le varibili Snmp coinvolte e le traduco nei corrispondenti OID
  		//numerici
  		//N.B. prevedo che cio' che mi viene restituito sia una lista degli OID non ripetuti (vengono eliminati gli eventuali duplicati)
  		String[] res = parseHF(HF);
  		if (res == null)
  			return null;

  		 HF = res[res.length-1];
  		 String oids[] = new String[res.length-1];
  		 for (int i=0; i<res.length-1; i++)
  		 	oids[i] = res[i];

   	
        int NP = hosts.length;
		Object result[] = new Object[NP];
		
   		int[] reqIDs = new int[NP];
   		SnmpSession[] sessions = new SnmpSession[NP];
   		//creo i ricevitori asincroni da associare ad ogni sessione
   		ReceiverAtomicReq[] receivers = new ReceiverAtomicReq[NP];
   		for(int j=0; j<NP; j++)
   		{
   			sessions[j] = new SnmpSession(api);
   			receivers[j] = new ReceiverAtomicReq(debug,out);
   		}
   		Settings[] sets = prepareAsyncSessions(sessions,hosts,ports,args,receivers,false);

        //qui invio tutte le richieste asincrone sulle varie sessioni
        SnmpPDU pdu;
        for (int k=0; k<NP; k++)
        {
        	//faccio la richiesta asincrona lasciando al sistema la scelta del reqID
        	int id = 0;
        	pdu = snmpAsyncOp(sets[k],sessions[k],oids,GET,null,id);
        	//caso in cui la get non sia possibile
        	if (pdu == null)
        		reqIDs[k] = -1;
			else
        		reqIDs[k] = pdu.getReqid();
        }

        SnmpPDU[] pdus = pickupAtomicAnswers(reqIDs,sessions,receivers);
        
        for (int x=0; x<pdus.length; x++)
        {
        	//ora recupero dalla pdu di risposta il valore di ogni var e lo associo all'oid corrispondente in una hashtable cosi' da
  			//poterli poi recuperare  facilmente al momento della valutazione
  			Hashtable values = new Hashtable();
  			int j =0;
  			for (Enumeration e = pdus[x].getVariableBindings().elements(); e.hasMoreElements(); )
  			{
  				SnmpVarBind varb = (SnmpVarBind) e.nextElement();
  				Object val = varb.getVariable().toValue();
  		System.out.println(oids[j]+" : "+val);
  				try
  				{  values.put(oids[j++],val);  }
  				catch (NullPointerException ex) { trace_err("Value of variable "+oids[j]+" not available");  return null;}
  			}

  			//chiamo la funzione di valutazione della HF passando come primo operando null, come numero di parentesi aperte 0 e
  			//trasmettendo la conoscenza delle associazioni variabile-valore
  			result[x] = evalHF(new ret(HF), values).ris;
  		}
  		
  		return result;

	}


/*VERSIONE SEQUENZIALE	
   public Object[] snmpHF(String[] hosts,int[] ports,String args[],String HF)
   {
    	int NP = hosts.length;
   		Vector ris = new Vector();
   		
   		for (int j=0; j<NP; j++)
   		{
   			args[Settings.PORT] = (new Integer(ports[j])).toString();
   			ris.add(snmpHF(hosts[j],args,HF,true));
   		}
   		
   		Object[] res = (Object[])ris.toArray(new Object[0]);
   		return res;
   	}*/

   public Object[] snmpHF(PlaceID[] places,int[] ports,String args[],String HF)
   {
	   	String[] hosts = new String[places.length];
	   	//recupero gli hosts corrispondenti ai places
		for (int j=0; j<places.length; j++)
            hosts[j] = conf.PlaceToHost(places[j]);

		return (snmpHF(hosts,ports,args,HF));
    }
    
 /*  public Object[] snmpMaxHF(String[] hosts,int[] ports,String args[],String HF)
   {
    	int NP = hosts.length;
    	Object max = null;
    	Vector where = new Vector();
   		
   		for (int j=0; j<NP; j++)
   		{
   			args[Settings.PORT] = (new Integer(ports[j])).toString();
   			Object cur = snmpHF(hosts[j],args,HF,true);
   			setMax(cur,max,where,ports[j]);
   		}
   		
   		Object[] ris = new Object[1];
   		ris[0] = max;
   		return ris;
   	}

   public Object[] snmpMaxHF(PlaceID[] places,int[] ports,String args[],String HF)
   {
	   	String[] hosts = new String[places.length];
	   	//recupero gli hosts corrispondenti ai places
		for (int j=0; j<places.length; j++)
            hosts[j] = conf.PlaceToHost(places[j]);

		return (snmpMaxHF(hosts,ports,args,HF));
	}

    void setMax (Object cur, Object max, Vector where,int port)
    {
    	int ret = -1;
   	 	if (cur instanceof Integer)
   	    	ret = ((Integer)cur).compareTo((Integer)max);
   		else if (cur instanceof Long)
 	    	ret = ((Long)cur).compareTo((Long)max);
   		else 
   			return;
   		
		if (ret > 0) 
		{
			max = cur;
			IpAddress iplocal = null;
			try{iplocal = new IpAddress(InetAddress.getLocalHost(),port);}
			catch(Exception e){ iplocal = new IpAddress("127.0.0.1",port);}
			where.removeAllElements();
			where.add(iplocal);
		}
		else if  (ret == 0)
		{
			IpAddress iplocal = null;
			try{iplocal = new IpAddress(InetAddress.getLocalHost(),port);}
			catch(Exception e){ iplocal = new IpAddress("127.0.0.1",port);}
			where.add(iplocal);
		}
		return ;
	}*/

    /*esegue la valutazione di una health function specificata dall'utente su ogni riga 
    //di una tabella SNMP, a partire dai parametri specificati in
   //input: ciascuno occupa una precisa posizione all'interno di args[]
   //host e' l'indirizzo presso cui si trova l'agente Snmp da interrogare
   //args e' il vettore dei parametri: la posizione di ogni parametro e' definita dalle variabili
   //static della classe Settings
   //HF e' l'health function sotto forma di stringa espressa in funzione di variabili Snmp in forma numerica e/o simbolica
   //from_config stabilisce se i parametri sono tutti contenuti in args (false) o se devono anche essere recuperati
   //dalla configurazione del motore Snmp locale (true)
   //restituisce l'Hashtable delle corrispondenze: identificatore di riga --> valore della HF sulla riga
   */
	public Hashtable snmpHFTable(String host, String args[], String HF, boolean from_config)
	{
        SnmpSession session = new SnmpSession(api);

        //se si richiede di ricorrere ai dati di configurazione si recuperano gli argomenti mancanti
    	//altrimenti e' gia' tutto in args
    	if (from_config)
    		if (prepareArguments(host,args) == false)
    			return null;

        //con tutti gli argomenti che ho preparo la sessione facendo gli opportuni settaggi
        Settings sets = prepareSession(host,args,from_config,session);
        
 		
  		return (snmpHFTable(sets,session,HF));
  	}
	
  	public Hashtable snmpHFTable(PlaceID place,String args[],String HF,boolean from_config)
  	{
       String host = conf.PlaceToHost(place);

       return(snmpHFTable(host,args,HF,from_config));
    }

	
	//idem ma su sessione gia' pronta
	public Hashtable snmpHFTable(Settings sets, SnmpSession session, String HF)
	{
		String[] res = parseHF(HF);
		HF = res[res.length-1];
		String oids[] = new String[res.length];
		
		//si ricava l'OID dell'identificatore di riga (il primo elemento della riga)
		//table.entry.index
		String oid_row = oids[0].substring(0,oids[0].length()-1)+"1";
		
		//creo la lista degli OID da recuperare:il primo e' l'identificatore di riga 
		//che serve per distinguere i risultati, gli altri sono quelli che servono per calcolare
		//l'HF 
		oids[0] = oid_row;
		for (int i=0; i<res.length; i++)
			oids[i] = res[i-1];
		
		//li copio in una variabile temporanea che alla ricezione di ogni risultato verra' 
		//modificata per fare la successiva richiesta	
		String[] tempOIDs = new String[oids.length];
		for (int j=0; j<oids.length;j++)
			tempOIDs[j] = oids[j];
		Hashtable rowsHF = new Hashtable();
		boolean next = true;
		while (next)
		{
			SnmpPDU ris = snmpOp(sets,session,tempOIDs,GETNEXT,null);
			//oidtab = oid riga -4
			//es oid tab = .1.3.6.1.2.1.2.2   oid riga = .1.3.6.1.2.1.2.2.1.1
			//controllo che si tratti ancora di valori di riga della tabella
			next = check(ris,oid_row.substring(0,oid_row.length()-4));
				
			if (next)
			{
				Object whichrow = null;
				Hashtable values = new Hashtable();
				int j=0;
				for (Enumeration e = ris.getVariableBindings().elements(); e.hasMoreElements();)
				{
					SnmpVarBind varb = (SnmpVarBind) e.nextElement();
					//la prima variabile e' il valore dell'identificatore di riga
					if (j == 0)
					{
						whichrow = varb.getVariable().toValue();
						j = 0;
					}
					else
					{
						Object val = varb.getVariable().toValue();
						try{ values.put(oids[j++],val);}
  						catch (NullPointerException ex) { trace_err("Value of variable "+tempOIDs[j]+" not available");  return null;}
  					}
  				}
  				
  				rowsHF.put(whichrow,evalHF(new ret(HF),values).ris);
  				
  				for (int i = 0; i<tempOIDs.length; i++)
  					tempOIDs[i] = ris.getObjectID(i).toString();
  			}
  		}
  		
  		return rowsHF;
  	}
  	
  	/**Recupera il valore di una HF su una tabella per piu' target
  	*  restituisce un array di Hashtable costituite dalle associazioni oidriga --> valore HF
  	*/
	public Hashtable[] snmpHFTable(String[] hosts,int[] ports,String args[],String HF)
	{
		int N = hosts.length;
		Hashtable[] res = new Hashtable[N];
		for (int j=0; j<N; j++)
		{
			args[Settings.PORT] = String.valueOf(ports[j]);
			res[j] = snmpHFTable(hosts[j],args,HF,true);
		}
		return res;
	}
			  	
   public Hashtable[] snmpHFTable(PlaceID[] places,int[] ports,String args[],String HF)
   {
	   	String[] hosts = new String[places.length];
	   	//recupero gli hosts corrispondenti ai places
		for (int j=0; j<places.length; j++)
            hosts[j] = conf.PlaceToHost(places[j]);

		return (snmpHFTable(hosts,ports,args,HF));
    }
    
    
  	/** Converte una stringa contenente una sequenza di parametri
 	 * in una Collection di singoli parametri.
  	*/
  	public Collection StringToParameters( String ParametersLine )
  	{
    		final String TokenDelimiters = " ";       // Elenco separatori di items: scartati nella lettura
    		Collection Parameters = new Vector();

    		StringTokenizer stok = new StringTokenizer( ParametersLine, TokenDelimiters , true );

    		try
    		{
      			while( stok.hasMoreTokens() )
      			{
        				String Token = stok.nextToken();

			        if( TokenDelimiters.indexOf( Token ) == -1 )  // Scarto gli spazi
         				 Parameters.add( Token );
      			}
    		}
    		catch( NoSuchElementException  e )
    		{}   // Nessun problema: semplicemente  finita la stringa!

    		//System.out.println( "Parameters: " + Parameters );
    		return Parameters;
 	 }
 	 
 	 
  	//restituisce la lista priva di ripetizioni degli OID numerici associati alle variabili contenute nella HF
  	//e la stringa stessa aggiornata con gli oid numerici al posto di quelli simbolici
	public String[] parseHF(String HF)
	{
		//escludendo operatori e parentesi gli altri sono tutti operandi
		//quindi corrispondono ad altrettante var Snmp (oid)
		StringTokenizer stHF = new StringTokenizer(HF," ()+-*/",false);

		//estraggo tutti i nomi delle variabili dalla stringa della hf e li inserisco in un hashset che mi da' garanzia di
		//non duplicazione di elementi
		HashSet nonrep = new HashSet();
		while (stHF.hasMoreTokens())
		{
			String temp = stHF.nextToken();
			nonrep.add(temp);
		}

		//trasformo ora in array
		String oids[] = (String[]) nonrep.toArray(new String[0]);

		//creo associazioni oid simbolico --> indice che individuera' il corrispondente numerico  nell'array degli oid numerci
		//mi servono, perche' dovro' anche inserire nella stringa i numerici
		Hashtable sym = new Hashtable();
		for (int y=0; y<oids.length; y++)
			sym.put(oids[y],new Integer(y));

		//converto anche da simbolico a numerico nel caso ce ne sia bisogno
		if  (conf.getNumericOIDs(oids) == false)
			return null;

		//devo anche inserire nella stringa della hf i numerici per poi riuscire a recuperarne i valori
		String tempHF = new String(HF);
		HF = new String("");
		StringTokenizer symHF = new StringTokenizer(tempHF," ()+-*/",true);

		while (symHF.hasMoreTokens())
		{
			String tok = symHF.nextToken();
			if (tok.equals(" "))
				;
			//se operatore o parentesi mi limito a ricopiarlo
			else if ( tok.equals("+") || tok.equals("*") || tok.equals("/") || tok.equals("-") || tok.equals("(") || tok.equals(")") )
				HF = HF+" "+tok;
			else //variabile => devo recuperare il numerico corrispondente
			{
				tok = oids[((Integer)sym.get(tok)).intValue()];
				HF = HF+" "+tok;
			}
		}

		//da sostituire con StringBuffer le cui modifiche possono sopravvivere a chiamata
		String dacambiare[] = new String[oids.length+1];
		int i=0;
		for (; i<oids.length; i++)
			dacambiare[i] = oids[i];
		dacambiare[i] = HF;
		return dacambiare;

	}

	//classe d'appoggio nell'implementazione della valutazione di una health function
	 class  ret
	 {
	 	//primo operando
	 	Object ris = null;
	 	//numero parentesi aperte ancora da bilanciare
	 	int nopen = 0;
	 	//collezione dei parametri costituenti l'hf ancora da considerare
	 	Collection rem = null;

	 	ret (String hf)
	 	{
	 		rem = StringToParameters(hf);
	 	}

	 	ret (Collection coll, Object op, int no)
	 	{
	 		rem = coll;
	 		ris = op;
	 		nopen = no;
	 	}
	 }

	ret evalHF(ret info, Hashtable values)
	{
		//recupero primo operando, numero parentesi aperte ancora da bilanciare, collezione di stringhe ancora da valutare
		Object op1 = info.ris;
		int open = info.nopen;
		Collection cHF = info.rem;
		ret next = null;

		//la Collection nella ricorsione dovrebbe risultare modificata
		Iterator iHF = cHF.iterator();

		String firstok = (String)iHF.next();
		iHF.remove();
		String temp = null;
	         if (firstok.equals("("))
		 {
		 	open ++;
		 	temp = (String)iHF.next();
		 	iHF.remove();
		 	//prima scarto tutte le eventuali parentesi aperte innestate e le conto (perche' dovranno essere bilanciate)
		 	while( temp .equals("(") && iHF.hasNext() )
		 	{
		 		open ++;
				temp = (String)iHF.next();
				iHF.remove();
			}
			//quando mi fermo e' perche' ho trovato un operando (variabile Snmp), devo recuperare il valore dalla hashtable
			Object  temp1 = values.get(temp);
			next = new ret(cHF,temp1,open);
			return (evalHF(next,values));
		}
		else if (firstok.equals("*") || firstok.equals("+") || firstok.equals("-") || firstok.equals("/") )
		{
			temp = (String)iHF.next();
		 	iHF.remove();

			Object op2 = null;
			if (temp.equals("(")) //se c'e' ( devo valutare un'altra espressione per ottenere il secondo operando
			         info  = evalHF(new ret(cHF,null,open++),values);
			else   //deve essere il secondo operando
				info = evalHF(new ret(cHF,values.get(temp),open),values);

			op2 = info.ris;
			open = info.nopen;
			cHF = info.rem;

			Object ris = compute(op1,firstok,op2);
			if (open != 0) //ci sono ancora parentesi da bilanciare => stringa non tutta consumata
				return (evalHF(new ret(cHF,ris,open),values));
			else
				return (new ret(cHF,ris,open));

		}
		else if ( firstok.equals(")") )
		{
			open--;
			if (iHF.hasNext())
			{
		 		temp = (String)iHF.next();
		 		iHF.remove();
		 	}
		 	else
		 		return  (new ret(cHF,op1,open));
		 	//prima scarto tutte le eventuali parentesi chiuse innestate e decremento il numero di parentesi aperte che devono
		 	// ancora essere  bilanciate
			while( temp .equals(")") && iHF.hasNext() )
		 	{
		 		open--;
				temp = (String)iHF.next();
				iHF.remove();
			}
			//devo reintrodurre l'ultimo elemento letto che non e' una parentesi chiusa per essere valutato in seguito
			iHF = cHF.iterator();

			//quando esco dal while precedente e' perche' sono finite le parentesi chiuse, restituisco come risultato l'operatore
			//che mi e' stato passato come argomento e che ora puo' essere utilizzato per il calcolo effettivo
			return (new ret(cHF,op1,open));
		}
		else  //c'e' un semplice operando (variabile) di cui mi limito arecuperare il valore
			if (iHF.hasNext()) //ci sara' un operatore
				return evalHF(new ret(cHF, values.get(firstok),open),values);
			return (new ret(cHF,values.get(firstok),open));
	}

	//calcola  "op1 optor op2" , dove op1 ed op2 sono valori che derivano da variabili Snmp, percio' si puo' trattare di Integer o Long
	 Object compute (Object op1,String optor,Object op2)
	 {
	 	if (op1 == null)
	 		return op2;
	 	if (op2 == null)
	 		return op1;

	 	if ( (op1 instanceof Long) && (op2 instanceof Long) )
	 	{
	 		long l1 = ( (Long) op1).longValue();
	 		long l2 = ( (Long) op2).longValue();

	 		if (optor.equals("+"))
	 			return ( new Long( l1+l2 ) );
	 		if (optor.equals("-"))
	 			return ( new Long( l1-l2  ) );
	 		if (optor.equals("*"))
	 			return ( new Long( l1*l2 ) );
	 		if (optor.equals("/"))
				return ( new Long( l1/l2  ) );
	 	}
		if ( (op1 instanceof Integer) && (op2 instanceof Integer)  )
	 	{
	 		int i1 = ((Integer) op1).intValue();
	 		int i2 = ((Integer) op2).intValue();

	 		if (optor.equals("+"))
	 			return ( new Integer( i1+i2  ) );
	 		if (optor.equals("-"))
	 			return ( new Integer( i1-i2  ) );
	 		if (optor.equals("*"))
	 			return ( new Integer( i1*i2  ) );
	 		if (optor.equals("/"))
				return ( new Integer( i1/i2 ) );
		}
	 	return null;
	 }

	
}
